
/*
  Copyright (C) 2003  Dmitry V. Levin <ldv@altlinux.org>

  The caller data initialization module for the pkg-build-priv program.

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA.
*/

#define _GNU_SOURCE
#include <errno.h>
#include <error.h>
#include <stdio.h>
#include <fcntl.h>
#include <paths.h>
#include <string.h>
#include <stdlib.h>
#include <signal.h>
#include <unistd.h>
#include <sys/wait.h>

extern const char *__progname;

static void
do_child (int ac, const char **av)
{
	const char path[] = "/etc/X11/prefdm";
	const char *args[(ac > 1) ? ac + 2 : 3];

	if (setsid () < 0)
		error (EXIT_FAILURE, errno, "setsid");

	args[0] = "prefdm";
	args[1] = "-nodaemon";
	if (ac > 0)
		memcpy (&args[2], &av[0], ac * sizeof (av[0]));
	args[2 + ac] = 0;

	execv (path, (char *const *)args);
	error (EXIT_FAILURE, errno, "%s", path);
}

static volatile int terminated, sigchld;

static void
term_handler (int no)
{
	terminated = 1;
}

static void
child_handler (int no)
{
	sigchld = 1;
}

int
main (int ac, const char **av)
{
	const char pid_file[] = "/var/run/rundm.pid";
	const char first_file[] = "/var/lock/TMP_1ST";
	FILE   *f;
	pid_t   pid;
	int i;

	if (ac < 1)
		return EXIT_FAILURE;
	--ac;
	++av;

	if (daemon (0, 0) < 0)
		error (EXIT_FAILURE, errno, "daemon");

	signal (SIGHUP, SIG_IGN);
	signal (SIGINT, term_handler);
	signal (SIGQUIT, term_handler);
	signal (SIGTERM, term_handler);
	signal (SIGCHLD, child_handler);

	f = fopen (pid_file, "w");
	if (f)
	{
		fprintf (f, "%ld\n", (long)getpid());
		fclose (f);
	}

	while (!terminated && !access (first_file, F_OK))
		sleep(1);

	if (terminated)
	{
		unlink (pid_file);
		return EXIT_SUCCESS;
	}

	for (i = 0; i < ac; ++i)
	{
		int fd, flags;

		if (!strcmp ("--", av[i]))
		{
			++i;
			break;
		}
		if ((fd = open (av[i], O_RDONLY|O_NOCTTY)) < 0)
		{
			error (EXIT_SUCCESS, errno, "open: %s", av[i]);
			continue;
		}
		if ((flags = fcntl (fd, F_GETFD, 0)) < 0)
		{
			error (EXIT_SUCCESS, errno, "fcntl: F_GETFD: %s", av[i]);
			close (fd);
			continue;
		}
		flags |= FD_CLOEXEC;
		if (fcntl (fd, F_SETFD, flags) < 0)
		{
			error (EXIT_SUCCESS, errno, "fcntl: F_SETFD: %s", av[i]);
			close (fd);
			continue;
		}
	}
	ac -= i;
	av += i;

	pid = fork ();
	if (pid < 0)
	{
		unlink (pid_file);
		error (EXIT_FAILURE, errno, "fork");
	}
	else if (!pid)
		do_child (ac, av);

	while (!terminated && !sigchld)
		pause ();

	if (terminated)
		kill (pid, SIGTERM);

	TEMP_FAILURE_RETRY (waitpid (pid, 0, 0));
	unlink (pid_file);

	return EXIT_SUCCESS;
}
